#include "X3D_CoordinateDeformer.h"
#include "X3D_X3DSphapeNode.h"

#include <iostream>

using namespace std;

namespace X3DTK {
namespace X3D {

CoordinateDeformer::CoordinateDeformer()
  : X3DGroupingNode(), X3DBoundedObject(),
    _controlPoint  (MFVec3f()),
    _inputCoord    (MFNode()),
    _inputTransform(MFNode()),
    _outputCoord   (MFNode()),
    _weight        (MFFloat()),
	 _uDimension    (0),
	 _uKnot         (MFDouble()),
	 _uOrder        (2),
	 _vDimension    (0),
	 _vKnot         (MFDouble()),
	 _vOrder        (2),
	 _wDimension    (0),
	 _wKnot         (MFDouble()),
	 _wOrder        (2)
{
  defineTypeName("CoordinateDeformer", "NURBS");
}

CoordinateDeformer::CoordinateDeformer(const MFVec2f &point)
  : X3DParametricGeometryNode(),
    _point(point)
{
  defineTypeName("CoordinateDeformer", "NURBS");
}

CoordinateDeformer::CoordinateDeformer(const CoordinateDeformer &N)
  : X3DParametricGeometryNode(N),
    _point(N._point)
{
}

SFAbstractNode CoordinateDeformer::clone() const
{
  return new CoordinateDeformer(*this);
}

CoordinateDeformer::~CoordinateDeformer()
{
}

bool Contour2D::addChild(const SFAbstractNode &N)
{
  if ( (dynamic_cast<NurbsCurve *>(N) != 0) ||
       (dynamic_cast<ContourPolyline2D *>(N) != 0) )
    return X3DGroupingNode::addChild(N);
  
  cx3d << "Contour2D::addChild : a node of type " << N->getTypeName() << " cannot be a child!" << endl;
  return false;    
}

bool Contour2D::setChild(const SFAbstractNode &N)
{
  if ( (dynamic_cast<NurbsCurve *>(N) != 0) ||
       (dynamic_cast<ContourPolyline2D *>(N) != 0) )
    return X3DGroupingNode::setChild(N);
  
  cx3d << "Contour2D::setChild : a node of type " << N->getTypeName() << " cannot be a child!" << endl;
  return false;    
}

void CoordinateDeformer::setPoint(const MFVec2f &point)
{
  _point = point;
}

void Cylinder::load(const X3DFileElement *element)
{
  int index;
  index = element->getIndexAttribute("point");
  if (index != -1)
    _point = MFVec2f(element->getAttribute(index));
}

SFString &Cylinder::write(SFString &output) const
{
  if (!_point.empty())
    output += " point=\"" + toSFString(_point) + "\"";
      
  return output;
}


}
}
