#include "MESH_ConnectedComComputer.h"
#include "MESH_ConnectedComComputerCoreVisitor.h"
#include "MESH_ConnectedComComputerMeshCoreVisitor.h"
#include "MESH_ConnectedComComputerStateVariables.h"
#include "DFSGraphTraversal.h"
#include "Progress.h"

#include <iostream>

using namespace std;

namespace X3DTK {
namespace MESH {

ConnectedComComputer::ConnectedComComputer()
: X3DMultiPassProcessor()
{
  progress = new Progress(256);
  Singleton<ConnectedComComputerStateVariables>::getInstance()->setProgress(progress);
  
  setGraphTraversal("Vertex", new DFSGraphTraversal());
  setComponentVisitor("Vertex", new ConnectedComComputerCoreVisitor(progress));
  
  setGraphTraversal("Mesh", new DFSGraphTraversal());
  setComponentVisitor("Mesh", new ConnectedComComputerMeshCoreVisitor());
}

ConnectedComComputer::~ConnectedComComputer()
{
  Singleton<ConnectedComComputerStateVariables>::removeInstance();
  delete progress;
}

void ConnectedComComputer::compute(Scene *S)
{
  compute(static_cast<X3DNode *>(S));
  
  S->getProcessHistory().addEntry("MESH::ConnectedComComputer");
}

void ConnectedComComputer::compute(X3DNode *N)
{
  Singleton<ConnectedComComputerStateVariables>::getInstance()->init();
  traverse("Vertex", N);
  Singleton<ConnectedComComputerStateVariables>::getInstance()->finish(); 

  Singleton<ConnectedComComputerStateVariables>::getInstance()->init();
  traverse("Mesh", N);
  Singleton<ConnectedComComputerStateVariables>::getInstance()->finish(); 
}

void ConnectedComComputer::setSelectedNodes(const MFNode &selectedNodes)
{
  Singleton<ConnectedComComputerStateVariables>::getInstance()->setSelectedNodes(selectedNodes);
}

}
}  
